% Get least-squares approximation for regularization term.
function [T, priorConfidenceWeights] = getRegularizationMatrix(imagePrior, x, M, N, priorConfidenceWeights, calmsrParams)  
    
    if isequal(imagePrior, 'TV')
        T = get_TV_regularization(x, M, N);
    elseif isequal(imagePrior, 'BTV')
        [T, priorConfidenceWeights] = get_BTV_regularization(x, M, N, priorConfidenceWeights, calmsrParams);
    else
        error(['Unknown image prior: ' imagePrior]);
    end

function T = get_TV_regularization(x, M, N)  
    
    % Derivative operators and weight matrices in horizontal and vertical
    % direction.
    [Vx, Wx] = compute_v_w(x, M, N, M, 1, 'TV');
    [Vy, Wy] = compute_v_w(x, M, N, 1, M, 'TV');
    
    % Assemble total variation matrix.
    Tx = Vx'*Wx*Vx;
    Ty = Vy'*Wy*Vy;
    T = Tx + Ty;

function [S, priorConfidenceWeights] = get_BTV_regularization(x, M, N, priorConfidenceWeights, calmsrParams)  

    shiftIdx = 1;    
    z = [];
    for l = -calmsrParams.btvWindowSize : calmsrParams.btvWindowSize
        % Get shift matrix for vertical direction.
        [Vl, Wl] = compute_v_w(x, M, N, M * l, l, 'BTV');
        for m = -calmsrParams.btvWindowSize : calmsrParams.btvWindowSize
             if l ~= 0 || m ~= 0
                 % Get shift matrix for horizontal direction
                [Vm, Wm] = compute_v_w(x, M, N, m, M * m, 'BTV');
                
                % Determine least-squares approximation of BTV for the
                % current shift.
                S_lm{shiftIdx} = Vl'*Wl*Vl + Vm'*Wm*Vm; 
                S_lm{shiftIdx} = calmsrParams.btvWeightingFactor^(abs(l) + abs(m)) .* S_lm{shiftIdx};
                
                if ~isempty(x)
                    % Determine filtered version of the BTV residual.
                    z{shiftIdx} = calmsrParams.btvWeightingFactor^(abs(l) + abs(m)) * (Vl + Vm) * x;
                    z{shiftIdx} = medfilt2(reshape(z{shiftIdx}, [M,N]), [3 3]);
                    z{shiftIdx} = z{shiftIdx}(:);
                end
                shiftIdx = shiftIdx + 1;
             end
        end
    end
    
    % Estimate prior weights and assemble overall BTV regularization
    % matrix.
    if ~isempty(calmsrParams.priorWeightingFunction) && ~isempty(z)
        z = cell2vec(z);
        [priorConfidenceWeights, ~] = calmsrParams.priorWeightingFunction.function(z(:), priorConfidenceWeights, calmsrParams.priorWeightingFunction.parameters{1:end});
    else
        priorConfidenceWeights = ones( ((2*calmsrParams.btvWindowSize + 1)^2 - 1) * M*N, 1);
    end
    S = sparse(1:M*N, 1:M*N, zeros(1,M*N), M*N, M*N);
    for shiftIdx = 1:(size(priorConfidenceWeights, 1)/ (M*N))
        A = sparse(1:M*N, 1:M*N, priorConfidenceWeights((shiftIdx-1) * (M*N) + 1:shiftIdx*(M*N)), M*N, M*N);
        S = S + (A * S_lm{shiftIdx});
    end

function [V, W] = compute_v_w(x, M, N, offset_w, offset_v, imagePrior)
    
    numHRPixel = M*N;
    ind = 1:numHRPixel;
    ind_off = mod(ind + offset_w, numHRPixel);
    ind_off(ind_off == 0) = numHRPixel;
    
    Id = sparse(1:numHRPixel, 1:numHRPixel, ones(1, numHRPixel), numHRPixel, numHRPixel);
    new_indx = mod(ind + offset_v, numHRPixel);
    new_indx(new_indx ==0) = numHRPixel;    
    V = sparse(ind', new_indx, -ones(1,numHRPixel), numHRPixel, numHRPixel);
    V = Id + V;
    
    if ~isempty(x)
        % Determine weight matrix for least-squares approximation of the
        % regularization term.
        e = abs(V*x); %abs( x(ind_off) - x(ind) );
        if isequal(imagePrior, 'TV')
            e = sqrt(e.^2 + 1e-4);
        else
            e(e < 1e-2) = 1e-2;
        end
        W = sparse(1:numHRPixel, 1:numHRPixel, 1 ./ e, numHRPixel, numHRPixel);
    else
        W = speye(numHRPixel, numHRPixel);
    end